@@ -21,6 +21,7 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 #include "gailbutton.h"
+#include <libgail-util/gailmisc.h>
 
 static void                  gail_button_class_init       (GailButtonClass *klass);
 static void                  gail_button_object_init      (GailButton      *button);
@@ -72,6 +73,49 @@ static void                  gail_button_get_image_size (AtkImage       *image,
 static gboolean              gail_button_set_image_description 
                                                         (AtkImage       *image,
                                                          const gchar    *description);
+
+/* atktext.h */ 
+static void	  atk_text_interface_init	   (AtkTextIface	*iface);
+
+static gchar*	  gail_button_get_text		   (AtkText	      *text,
+                                                    gint	      start_pos,
+						    gint	      end_pos);
+static gunichar	  gail_button_get_character_at_offset(AtkText	      *text,
+						    gint	      offset);
+static gchar*     gail_button_get_text_before_offset(AtkText	      *text,
+ 						    gint	      offset,
+						    AtkTextBoundary   boundary_type,
+						    gint	      *start_offset,
+						    gint	      *end_offset);
+static gchar*     gail_button_get_text_at_offset   (AtkText	      *text,
+ 						    gint	      offset,
+						    AtkTextBoundary   boundary_type,
+						    gint	      *start_offset,
+						    gint	      *end_offset);
+static gchar*     gail_button_get_text_after_offset(AtkText	      *text,
+ 						    gint	      offset,
+						    AtkTextBoundary   boundary_type,
+						    gint	      *start_offset,
+						    gint	      *end_offset);
+static gint	  gail_button_get_character_count  (AtkText	      *text);
+static void gail_button_get_character_extents      (AtkText	      *text,
+						    gint 	      offset,
+		                                    gint 	      *x,
+                    		   	            gint 	      *y,
+                                		    gint 	      *width,
+                                     		    gint 	      *height,
+			        		    AtkCoordType      coords);
+static gint gail_button_get_offset_at_point        (AtkText           *text,
+                                                    gint              x,
+                                                    gint              y,
+			                            AtkCoordType      coords);
+static AtkAttributeSet* gail_button_get_run_attributes 
+                                                   (AtkText           *text,
+              					    gint 	      offset,
+                                                    gint 	      *start_offset,
+					            gint	      *end_offset);
+static AtkAttributeSet* gail_button_get_default_attributes
+                                                   (AtkText           *text);
 static GtkImage*             get_image_from_button      (GtkWidget      *button);
 static GtkWidget*            get_label_from_button      (GtkWidget      *button);
 static void                  set_role_for_button        (AtkObject      *accessible,
@@ -114,6 +158,13 @@ gail_button_get_type (void)
         NULL
       };
 
+      static const GInterfaceInfo atk_text_info =
+      {
+        (GInterfaceInitFunc) atk_text_interface_init,
+        (GInterfaceFinalizeFunc) NULL,
+        NULL
+      };
+
       type = g_type_register_static (GAIL_TYPE_CONTAINER,
                                      "GailButton", &tinfo, 0);
 
@@ -121,6 +172,9 @@ gail_button_get_type (void)
                                    &atk_action_info);
       g_type_add_interface_static (type, ATK_TYPE_IMAGE,
                                    &atk_image_info);
+      g_type_add_interface_static (type, ATK_TYPE_TEXT,
+                                   &atk_text_info);
+
     }
 
   return type;
@@ -216,6 +270,8 @@ gail_button_real_initialize (AtkObject *obj,
 {
   GailButton *button = GAIL_BUTTON (obj);
   guint handler_id;
+  GtkWidget  *label;
+  const gchar *label_text;
 
   ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
@@ -246,6 +302,15 @@ gail_button_real_initialize (AtkObject *obj,
                     "leave",
                     G_CALLBACK (gail_button_released_leave_handler),
                     NULL);
+
+  button->textutil = gail_text_util_new ();
+
+  label = get_label_from_button (GTK_WIDGET (data));
+  if (GTK_IS_LABEL (label))
+    {
+      label_text = gtk_label_get_text (GTK_LABEL (label));
+      gail_text_util_text_setup (button->textutil, label_text);
+    }
 }
 
 static void
@@ -775,6 +840,325 @@ gail_button_set_image_description (AtkImage    *image,
     return FALSE;
 }
 
+/* atktext.h */
+
+static void
+atk_text_interface_init (AtkTextIface *iface)
+{
+  g_return_if_fail (iface != NULL);
+  iface->get_text = gail_button_get_text;
+  iface->get_character_at_offset = gail_button_get_character_at_offset;
+  iface->get_text_before_offset = gail_button_get_text_before_offset;
+  iface->get_text_at_offset = gail_button_get_text_at_offset;
+  iface->get_text_after_offset = gail_button_get_text_after_offset;
+  iface->get_character_count = gail_button_get_character_count;
+  iface->get_character_extents = gail_button_get_character_extents;
+  iface->get_offset_at_point = gail_button_get_offset_at_point;
+  iface->get_run_attributes = gail_button_get_run_attributes;
+  iface->get_default_attributes = gail_button_get_default_attributes;
+}
+
+static gchar*
+gail_button_get_text (AtkText *text,
+                      gint    start_pos,
+                      gint    end_pos)
+{
+  GtkWidget *widget;
+  GtkWidget  *label;
+
+  const gchar *label_text;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  if (widget == NULL)
+    /* State is defunct */
+    return NULL;
+
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL (label))
+    return NULL;
+
+  label_text = gtk_label_get_text (GTK_LABEL (label));
+ 
+  if (label_text == NULL)
+    return NULL;
+  else
+  {
+    return gail_text_util_get_substring (GAIL_BUTTON(text)->textutil, 
+                                         start_pos, end_pos);
+  }
+}
+
+static gchar*
+gail_button_get_text_before_offset (AtkText         *text,
+				    gint            offset,
+				    AtkTextBoundary boundary_type,
+				    gint            *start_offset,
+				    gint            *end_offset)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+  
+  widget = GTK_ACCESSIBLE (text)->widget;
+  
+  if (widget == NULL)
+    /* State is defunct */
+    return NULL;
+  
+  /* Get label */
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return NULL;
+
+  return gail_text_util_get_text (GAIL_BUTTON (text)->textutil,
+                           gtk_label_get_layout (GTK_LABEL (label)), GAIL_BEFORE_OFFSET, 
+                           boundary_type, offset, start_offset, end_offset); 
+}
+
+static gchar*
+gail_button_get_text_at_offset (AtkText         *text,
+			        gint            offset,
+			        AtkTextBoundary boundary_type,
+ 			        gint            *start_offset,
+			        gint            *end_offset)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+ 
+  widget = GTK_ACCESSIBLE (text)->widget;
+  
+  if (widget == NULL)
+    /* State is defunct */
+    return NULL;
+  
+  /* Get label */
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return NULL;
+
+  return gail_text_util_get_text (GAIL_BUTTON (text)->textutil,
+                              gtk_label_get_layout (GTK_LABEL (label)), GAIL_AT_OFFSET, 
+                              boundary_type, offset, start_offset, end_offset);
+}
+
+static gchar*
+gail_button_get_text_after_offset (AtkText         *text,
+				   gint            offset,
+				   AtkTextBoundary boundary_type,
+				   gint            *start_offset,
+				   gint            *end_offset)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  
+  if (widget == NULL)
+  {
+    /* State is defunct */
+    return NULL;
+  }
+  
+  /* Get label */
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return NULL;
+
+  return gail_text_util_get_text (GAIL_BUTTON (text)->textutil,
+                           gtk_label_get_layout (GTK_LABEL (label)), GAIL_AFTER_OFFSET, 
+                           boundary_type, offset, start_offset, end_offset);
+}
+
+static gint
+gail_button_get_character_count (AtkText *text)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  if (widget == NULL)
+    /* State is defunct */
+    return 0;
+
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return 0;
+
+  return g_utf8_strlen (gtk_label_get_text (GTK_LABEL (label)), -1);
+}
+
+static void
+gail_button_get_character_extents (AtkText      *text,
+				   gint         offset,
+		                   gint         *x,
+                    		   gint 	*y,
+                                   gint 	*width,
+                                   gint 	*height,
+			           AtkCoordType coords)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+  PangoRectangle char_rect;
+  gint index, x_layout, y_layout;
+  const gchar *label_text;
+ 
+  widget = GTK_ACCESSIBLE (text)->widget;
+
+  if (widget == NULL)
+    /* State is defunct */
+    return;
+
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return;
+  
+  gtk_label_get_layout_offsets (GTK_LABEL (label), &x_layout, &y_layout);
+  label_text = gtk_label_get_text (GTK_LABEL (label));
+  index = g_utf8_offset_to_pointer (label_text, offset) - label_text;
+  pango_layout_index_to_pos (gtk_label_get_layout (GTK_LABEL (label)), index, &char_rect);
+  
+  gail_misc_get_extents_from_pango_rectangle (label, &char_rect, 
+                    x_layout, y_layout, x, y, width, height, coords);
+} 
+
+static gint 
+gail_button_get_offset_at_point (AtkText      *text,
+                                 gint         x,
+                                 gint         y,
+			         AtkCoordType coords)
+{ 
+  GtkWidget *widget;
+  GtkWidget *label;
+  gint index, x_layout, y_layout;
+  const gchar *label_text;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  if (widget == NULL)
+    /* State is defunct */
+    return -1;
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return -1;
+  
+  gtk_label_get_layout_offsets (GTK_LABEL (label), &x_layout, &y_layout);
+  
+  index = gail_misc_get_index_at_point_in_layout (label, 
+                                              gtk_label_get_layout (GTK_LABEL (label)), 
+                                              x_layout, y_layout, x, y, coords);
+  label_text = gtk_label_get_text (GTK_LABEL (label));
+  if (index == -1)
+    {
+      if (coords == ATK_XY_WINDOW || coords == ATK_XY_SCREEN)
+        return g_utf8_strlen (label_text, -1);
+
+      return index;  
+    }
+  else
+    return g_utf8_pointer_to_offset (label_text, label_text + index);  
+}
+
+static AtkAttributeSet*
+gail_button_get_run_attributes (AtkText        *text,
+                                gint 	      offset,
+                                gint 	      *start_offset,
+	                        gint	      *end_offset)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+  AtkAttributeSet *at_set = NULL;
+  GtkJustification justify;
+  GtkTextDirection dir;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  if (widget == NULL)
+    /* State is defunct */
+    return NULL;
+
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return NULL;
+  
+  /* Get values set for entire label, if any */
+  justify = gtk_label_get_justify (GTK_LABEL (label));
+  if (justify != GTK_JUSTIFY_CENTER)
+    {
+      at_set = gail_misc_add_attribute (at_set, 
+                                        ATK_TEXT_ATTR_JUSTIFICATION,
+     g_strdup (atk_text_attribute_get_value (ATK_TEXT_ATTR_JUSTIFICATION, justify)));
+    }
+  dir = gtk_widget_get_direction (label);
+  if (dir == GTK_TEXT_DIR_RTL)
+    {
+      at_set = gail_misc_add_attribute (at_set, 
+                                        ATK_TEXT_ATTR_DIRECTION,
+     g_strdup (atk_text_attribute_get_value (ATK_TEXT_ATTR_DIRECTION, dir)));
+    }
+
+  at_set = gail_misc_layout_get_run_attributes (at_set,
+                                                gtk_label_get_layout (GTK_LABEL (label)),
+                                                (gchar *) gtk_label_get_text (GTK_LABEL (label)),
+                                                offset,
+                                                start_offset,
+                                                end_offset);
+  return at_set;
+}
+
+static AtkAttributeSet*
+gail_button_get_default_attributes (AtkText        *text)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+  AtkAttributeSet *at_set = NULL;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  if (widget == NULL)
+    /* State is defunct */
+    return NULL;
+
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return NULL;
+
+  at_set = gail_misc_get_default_attributes (at_set,
+                                             gtk_label_get_layout (GTK_LABEL (label)),
+                                             widget);
+  return at_set;
+}
+
+static gunichar 
+gail_button_get_character_at_offset (AtkText	         *text,
+                                     gint	         offset)
+{
+  GtkWidget *widget;
+  GtkWidget *label;
+  const gchar *string;
+  gchar *index;
+
+  widget = GTK_ACCESSIBLE (text)->widget;
+  if (widget == NULL)
+    /* State is defunct */
+    return '\0';
+
+  label = get_label_from_button (widget);
+
+  if (!GTK_IS_LABEL(label))
+    return '\0';
+  string = gtk_label_get_text (GTK_LABEL (label));
+  if (offset >= g_utf8_strlen (string, -1))
+    return '\0';
+  index = g_utf8_offset_to_pointer (string, offset);
+
+  return g_utf8_get_char (index);
+}
+
 static void
 gail_button_finalize (GObject            *object)
 {
