@@ -166,8 +166,8 @@ gail_label_real_initialize (AtkObject *obj,
   
   gail_label = GAIL_LABEL (obj);
 
-  gail_label->cursor_position = -1;
-  gail_label->selection_bound = -1;
+  gail_label->cursor_position = 0;
+  gail_label->selection_bound = 0;
   
   gail_label->textutil = gail_text_util_new ();
 
@@ -277,8 +277,8 @@ gail_label_real_notify_gtk (GObject           *obj,
             {
               /* GtkLabel has become non selectable */
 
-              gail_label->cursor_position = -1;
-              gail_label->selection_bound = -1;
+              gail_label->cursor_position = 0;
+              gail_label->selection_bound = 0;
               text_caret_moved = TRUE;
             }
             
