@@ -149,8 +149,10 @@ static gchar*            get_text_near_offset          (AtkText          *text,
                                                         gint             *start_offset,
                                                         gint             *end_offset);
 static gint             get_insert_offset              (GtkTextBuffer    *buffer);
-static void             emit_text_caret_moved          (AtkObject        *obj,
+static gint             get_selection_bound            (GtkTextBuffer    *buffer);
+static void             emit_text_caret_moved          (GailTextView     *gail_text_view,
                                                         gint             insert_offset);
+static gint             insert_idle_handler            (gpointer         data);
 
 static GailWidgetClass *parent_class = NULL;
 
@@ -233,6 +235,8 @@ gail_text_view_init (GailTextView      *text_view)
   text_view->textutil = NULL;
   text_view->signal_name = NULL;
   text_view->previous_insert_offset = -1;
+  text_view->previous_selection_bound = -1;
+  text_view->insert_notify_handler = 0;
 }
 
 static void
@@ -269,6 +273,9 @@ gail_text_view_finalize (GObject            *object)
   GailTextView *text_view = GAIL_TEXT_VIEW (object);
 
   g_object_unref (text_view->textutil);
+  if (text_view->insert_notify_handler)
+    gtk_idle_remove (text_view->insert_notify_handler);
+
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -296,14 +303,13 @@ gail_text_view_real_notify_gtk (GObject             *obj,
 {
   if (!strcmp (pspec->name, "editable"))
     {
-      GValue value = { 0, };
       AtkObject *atk_obj;
+      gboolean editable;
 
-      g_value_init (&value, G_TYPE_BOOLEAN);
-      g_object_get_property (obj, "editable", &value);
       atk_obj = gtk_widget_get_accessible (GTK_WIDGET (obj));
+      editable = gtk_text_view_get_editable (GTK_TEXT_VIEW (obj));
       atk_object_notify_state_change (atk_obj, ATK_STATE_EDITABLE,
-                                      g_value_get_boolean (&value));
+                                      editable);
     }
   else
     parent_class->notify_gtk (obj, pspec);
@@ -455,6 +461,9 @@ gail_text_view_get_character_at_offset (AtkText *text,
     return '\0';
 
   buffer = GAIL_TEXT_VIEW (text)->textutil->buffer;
+  if (offset >= gtk_text_buffer_get_char_count (buffer))
+    return '\0';
+
   gtk_text_buffer_get_start_iter (buffer, &start);
   gtk_text_buffer_get_end_iter (buffer, &end);
   string = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
@@ -515,7 +524,7 @@ gail_text_view_set_caret_offset (AtkText *text,
   buffer = view->buffer;
 
   gtk_text_buffer_get_iter_at_offset (buffer,  &pos_itr, offset);
-  gtk_text_buffer_move_mark_by_name (buffer, "insert", &pos_itr);
+  gtk_text_buffer_place_cursor (buffer, &pos_itr);
   return TRUE;
 }
 
@@ -531,6 +540,7 @@ gail_text_view_get_offset_at_point (AtkText      *text,
   gint x_widget, y_widget, x_window, y_window, buff_x, buff_y;
   GtkWidget *widget;
   GdkWindow *window;
+  GdkRectangle rect;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -561,6 +571,13 @@ gail_text_view_get_offset_at_point (AtkText      *text,
 
   gtk_text_view_window_to_buffer_coords (view, GTK_TEXT_WINDOW_WIDGET,
                                          x, y, &buff_x, &buff_y);
+  gtk_text_view_get_visible_rect (view, &rect);
+  /*
+   * Clamp point to visible rectangle
+   */
+  buff_x = CLAMP (buff_x, rect.x, rect.x + rect.width - 1);
+  buff_y = CLAMP (buff_y, rect.y, rect.y + rect.height - 1);
+
   gtk_text_view_get_iter_at_location (view, &loc_itr, buff_x, buff_y);
   return gtk_text_iter_get_offset (&loc_itr);
 }
@@ -1370,13 +1387,42 @@ _gail_text_view_insert_text_cb (GtkTextBuffer *buffer,
   GtkTextView *text = (GtkTextView *) user_data;
   AtkObject *accessible;
   GailTextView *gail_text_view;
+  gint position;
+  gint length;
+
+  g_return_if_fail (arg3 > 0);
 
   accessible = gtk_widget_get_accessible(GTK_WIDGET(text));
   gail_text_view = GAIL_TEXT_VIEW (accessible);
 
   gail_text_view->signal_name = "text_changed::insert";
-  gail_text_view->position = gtk_text_iter_get_offset (arg1);
-  gail_text_view->length = arg3;
+  position = gtk_text_iter_get_offset (arg1);
+  length = arg3;
+  
+  if (gail_text_view->length == 0)
+    {
+      gail_text_view->position = position;
+      gail_text_view->length = length;
+    }
+  else if (gail_text_view->position + gail_text_view->length == position)
+    {
+      gail_text_view->length += length;
+    }
+  else
+    {
+      /*
+       * We have a non-contiguous insert so report what we have
+       */
+      if (gail_text_view->insert_notify_handler)
+        {
+          gtk_idle_remove (gail_text_view->insert_notify_handler);
+        }
+      gail_text_view->insert_notify_handler = 0;
+      insert_idle_handler (gail_text_view);
+      gail_text_view->position = position;
+      gail_text_view->length = length;
+    }
+    
   /*
    * The signal will be emitted when the changed signal is received
    */
@@ -1401,6 +1447,24 @@ _gail_text_view_delete_range_cb (GtkTextBuffer *buffer,
 
   accessible = gtk_widget_get_accessible(GTK_WIDGET(text));
   gail_text_view = GAIL_TEXT_VIEW (accessible);
+  if (gail_text_view->insert_notify_handler)
+    {
+      gtk_idle_remove (gail_text_view->insert_notify_handler);
+      gail_text_view->insert_notify_handler = 0;
+      if (gail_text_view->position == offset && 
+          gail_text_view->length == length)
+        {
+        /*
+         * Do not bother with insert and delete notifications
+         */
+          gail_text_view->signal_name = NULL;
+          gail_text_view->position = 0;
+          gail_text_view->length = 0;
+          return;
+        }
+
+      insert_idle_handler (gail_text_view);
+    }
   g_signal_emit_by_name (accessible, "text_changed::delete",
                          offset, length);
 }
@@ -1416,9 +1480,11 @@ _gail_text_view_mark_set_cb (GtkTextBuffer *buffer,
 {
   GtkTextView *text = (GtkTextView *) user_data;
   AtkObject *accessible;
+  GailTextView *gail_text_view;
   const char *mark_name = gtk_text_mark_get_name(arg2);
 
   accessible = gtk_widget_get_accessible(GTK_WIDGET(text));
+  gail_text_view = GAIL_TEXT_VIEW (accessible);
 
   /*
    * Only generate the signal for the "insert" mark, which
@@ -1426,21 +1492,41 @@ _gail_text_view_mark_set_cb (GtkTextBuffer *buffer,
    */
   if (mark_name && !strcmp(mark_name, "insert"))
     {
-      GtkTextIter selection_bound_iter;
-      int insert_offset, selection_bound_offset;
+      int insert_offset, selection_bound;
+      gboolean selection_changed;
 
       insert_offset = gtk_text_iter_get_offset (arg1);
-      emit_text_caret_moved (accessible, insert_offset);
 
-      gtk_text_buffer_get_iter_at_mark (buffer, &selection_bound_iter,
-                 gtk_text_buffer_get_mark (buffer, "selection_bound"));
-      selection_bound_offset = gtk_text_iter_get_offset (&selection_bound_iter);
+      selection_bound = get_selection_bound (buffer);
+      if (selection_bound != insert_offset)
+        {
+          if (selection_bound != gail_text_view->previous_selection_bound ||
+              insert_offset != gail_text_view->previous_insert_offset)
+            {
+              selection_changed = TRUE;
+            }
+          else
+            {
+              selection_changed = FALSE;
+            }
+        }
+      else if (gail_text_view->previous_selection_bound != gail_text_view->previous_insert_offset)
+        {
+          selection_changed = TRUE;
+        }
+      else
+        {
+          selection_changed = FALSE;
+        }
+
+      emit_text_caret_moved (gail_text_view, insert_offset);
       /*
        * insert and selection_bound marks are different to a selection
        * has changed
        */
-      if (selection_bound_offset != insert_offset)
+      if (selection_changed)
         g_signal_emit_by_name (accessible, "text_selection_changed");
+      gail_text_view->previous_selection_bound = selection_bound;
     }
 }
 
@@ -1456,13 +1542,14 @@ _gail_text_view_changed_cb (GtkTextBuffer *buffer,
   gail_text_view = GAIL_TEXT_VIEW (accessible);
   if (gail_text_view->signal_name)
     {
-      g_signal_emit_by_name (accessible,
-                             gail_text_view->signal_name,
-                             gail_text_view->position,
-                             gail_text_view->length);
-      gail_text_view->signal_name = NULL;
+      if (!gail_text_view->insert_notify_handler)
+        {
+          gail_text_view->insert_notify_handler = gtk_idle_add (insert_idle_handler, accessible);
+        }
+      return;
     }
-  emit_text_caret_moved (accessible, get_insert_offset (buffer));
+  emit_text_caret_moved (gail_text_view, get_insert_offset (buffer));
+  gail_text_view->previous_selection_bound = get_selection_bound (buffer);
 }
 
 static gchar*
@@ -1479,13 +1566,11 @@ get_text_near_offset (AtkText          *text,
   view = GTK_TEXT_VIEW (GTK_ACCESSIBLE (text)->widget);
 
   /*
-   * If wrapping is on for the GtkTextView, the GtkTextView is passed
-   * to the function gail_text_util_get_text() so it can find the start
-   * and end of the current line on the display.
+   * Pass the GtkTextView to the function gail_text_util_get_text() 
+   * so it can find the start and end of the current line on the display.
    */
-  if (view->wrap_mode != GTK_WRAP_NONE &&
-      (boundary_type == ATK_TEXT_BOUNDARY_LINE_START ||
-       boundary_type == ATK_TEXT_BOUNDARY_LINE_END))
+  if (boundary_type == ATK_TEXT_BOUNDARY_LINE_START ||
+      boundary_type == ATK_TEXT_BOUNDARY_LINE_END)
     layout = view;
 
   return gail_text_util_get_text (GAIL_TEXT_VIEW (text)->textutil, layout,
@@ -1504,13 +1589,30 @@ get_insert_offset (GtkTextBuffer *buffer)
   return gtk_text_iter_get_offset (&cursor_itr);
 }
 
-static void
-emit_text_caret_moved (AtkObject *obj,
-                       gint      insert_offset)
+static gint
+get_selection_bound (GtkTextBuffer *buffer)
 {
-  GailTextView *gail_text_view;
+  GtkTextMark *selection_mark;
+  GtkTextIter selection_itr;
 
-  gail_text_view = GAIL_TEXT_VIEW (obj);
+  selection_mark = gtk_text_buffer_get_selection_bound (buffer);
+  gtk_text_buffer_get_iter_at_mark (buffer, &selection_itr, selection_mark);
+  return gtk_text_iter_get_offset (&selection_itr);
+}
+
+static void
+emit_text_caret_moved (GailTextView *gail_text_view,
+                       gint          insert_offset)
+{
+  /*
+   * If we have text which has been inserted notify the user
+   */
+  if (gail_text_view->insert_notify_handler)
+    {
+      gtk_idle_remove (gail_text_view->insert_notify_handler);
+      gail_text_view->insert_notify_handler = 0;
+      insert_idle_handler (gail_text_view);
+    }
 
   if (insert_offset != gail_text_view->previous_insert_offset)
     {
@@ -1520,7 +1622,36 @@ emit_text_caret_moved (AtkObject *obj,
        * When mouse click is used to change caret position, notification
        * is received on button down and button up.
        */
-      g_signal_emit_by_name (obj, "text_caret_moved", insert_offset);
+      g_signal_emit_by_name (gail_text_view, "text_caret_moved", insert_offset);
       gail_text_view->previous_insert_offset = insert_offset;
     }
 }
+
+static gint
+insert_idle_handler (gpointer data)
+{
+  GailTextView *gail_text_view;
+  GtkTextBuffer *buffer;
+
+  gail_text_view = GAIL_TEXT_VIEW (data);
+
+  g_signal_emit_by_name (data,
+                         gail_text_view->signal_name,
+                         gail_text_view->position,
+                         gail_text_view->length);
+  gail_text_view->signal_name = NULL;
+  gail_text_view->position = 0;
+  gail_text_view->length = 0;
+
+  buffer = gail_text_view->textutil->buffer;
+  if (gail_text_view->insert_notify_handler)
+    {
+    /*
+     * If called from idle handler notify caret moved
+     */
+      gail_text_view->insert_notify_handler = 0;
+      emit_text_caret_moved (gail_text_view, get_insert_offset (buffer));
+      gail_text_view->previous_selection_bound = get_selection_bound (buffer);
+    }
+  return FALSE;
+}
